/*  GuiTuner 0.05
 *  Copyright (c) 2001 Fabio Checconi <fchecconi@libero.it>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifdef HAVE_LIBASOUND

#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <cstring>
#include <string>
#include "alsa.h"
#include "i18n.h"

bool
Tuner::Alsa::setParams(AudioInfo *ai)
{
	snd_pcm_channel_params_t params = convInfo(ai);
	if(ai->mode == PLAYBACK || ai->mode == DUPLEX) {
		params.channel = SND_PCM_CHANNEL_PLAYBACK;
		params.start_mode = SND_PCM_START_FULL;
		snd_pcm_plugin_flush(handle, SND_PCM_CHANNEL_PLAYBACK);
		if((lastError = snd_pcm_plugin_params(handle, &params)) < 0)
			return false;
		if((lastError = snd_pcm_plugin_prepare(handle, SND_PCM_CHANNEL_PLAYBACK)) <0)
			return false;
	}
	if(ai->mode == RECORD || ai->mode == DUPLEX) {
		params.channel = SND_PCM_CHANNEL_CAPTURE;
		params.start_mode = SND_PCM_START_DATA;
		params.stop_mode = SND_PCM_STOP_ROLLOVER;
		snd_pcm_plugin_flush(handle, SND_PCM_CHANNEL_CAPTURE);
		if((lastError = snd_pcm_plugin_params(handle, &params)) < 0)
			return false;
		if((lastError = snd_pcm_plugin_prepare(handle, SND_PCM_CHANNEL_CAPTURE)) <0)
			return false;
	}

	return true;
}

bool
Tuner::Alsa::getParams(AudioInfo *ai)
{
	snd_pcm_channel_setup_t setup;
	memset(&setup, 0, sizeof(setup));
	setup.mode = SND_PCM_MODE_BLOCK;

	if(info.mode == PLAYBACK || info.mode == DUPLEX)
		setup.channel = SND_PCM_CHANNEL_PLAYBACK;
	else
		setup.channel = SND_PCM_CHANNEL_CAPTURE;

	if((lastError = snd_pcm_plugin_setup(handle, &setup)) < 0)
		return false;

	if(!convParams(&setup, ai))
		throwError(std::string(_("Tuner::Alsa::openDevice")),
			Error::LIBRARY_CALL,
			std::string(_("Format not supported")));

	return true;
}

bool
Tuner::Alsa::convParams(snd_pcm_channel_setup_t *setup, AudioInfo *ai)
{
	switch(setup->format.format) {
		case SND_PCM_SFMT_U8:
			ai->fmt = U8;
			break;
		case SND_PCM_SFMT_S16_BE:
			ai->fmt = S16BE;
			break;
		case SND_PCM_SFMT_S16_LE:
			ai->fmt = S16LE;
			break;
		default:
			return false;
	}
	switch(setup->format.voices) {
		case 1:
			ai->chn = MONO;
			break;
		case 2:
			ai->chn = STEREO;
			break;
		default:
			return false;
	}
	ai->rate = setup->format.rate;
	ai->fragNumber = setup->buf.block.frags;
	ai->fragSize = setup->buf.block.frag_size;

	return true;
}

const snd_pcm_channel_params_t
Tuner::Alsa::convInfo(AudioInfo *ai)
{
	snd_pcm_channel_params_t params;

	std::memset(&params, 0, sizeof(params));
	params.mode = SND_PCM_MODE_BLOCK;
	params.stop_mode = SND_PCM_STOP_STOP;
	params.format.interleave = 1;
	switch(ai->fmt) {
		case U8:
			params.format.format = SND_PCM_SFMT_U8;
			break;
		case S16BE:
			params.format.format = SND_PCM_SFMT_S16_BE;
			break;
		case S16LE:
			params.format.format = SND_PCM_SFMT_S16_LE;
	}
	params.format.rate = ai->rate;
	params.format.voices = ai->chn == MONO ? 1: 2;
	params.buf.block.frag_size = setBuf ? ai->fragSize: 128;
	params.buf.block.frags_max = 1;
	params.buf.block.frags_min = 1;

	return params;
}

void
Tuner::Alsa::throwError(std::string s1, Error::AudioErrorID id, std::string s2)
	throw(Error::AudioError)
{
	if(open)
		snd_pcm_close(handle), handle = 0, open = false;
	throw Error::AudioError(s1, id, s2);
}

Tuner::Alsa::Alsa() throw(Error::AudioError): handle(0), open(false), setBuf(false),
	lastError(0)
{
	info.loadDefaults();
}

Tuner::Alsa::~Alsa()
{
	if(open)
		snd_pcm_close(handle), handle = 0, open = false;
}

bool
Tuner::Alsa::openDevice(AudioInfo *ai, bool block, bool setb)
	throw(Error::AudioError)
{
	int flags = block ? 0: SND_PCM_OPEN_NONBLOCK;
	switch(info.mode = ai->mode) {
		case PLAYBACK:
			flags |= SND_PCM_OPEN_PLAYBACK;
			break;
		case RECORD:
			flags |= SND_PCM_OPEN_CAPTURE;
			break;
		case DUPLEX:
			flags |= SND_PCM_OPEN_DUPLEX;
	}

	if((lastError = snd_pcm_open(&handle, 0, 0, flags)) < 0) {
		if(lastError == -EBUSY)
			return false;
		throwError(std::string(_("Tuner::Alsa::openDevice")),
			Error::LIBRARY_CALL,
			std::string(snd_strerror(lastError)));
	}

	open = true;

	snd_pcm_nonblock_mode(handle, 0);

	setBuf = setb;
	if(!setParams(ai))
		throwError(std::string(_("Tuner::Alsa::openDevice")),
			Error::LIBRARY_CALL,
			std::string(snd_strerror(lastError)));

	if(!getParams(ai))
		throwError(std::string(_("Tuner::Alsa::openDevice")),
			Error::LIBRARY_CALL,
			std::string(snd_strerror(lastError)));

	info = *ai;

	return true;
}

bool
Tuner::Alsa::closeDevice() throw()
{
	snd_pcm_close(handle), handle = 0, open = false;

	return true;
}

bool
Tuner::Alsa::play(unsigned char *buf, int len) throw(Error::AudioError)
{
	int count = len;
	while(count >= info.fragSize) {
		int val = snd_pcm_plugin_write(handle, buf + len - count, info.fragSize);
		if(val <= 0)
			return false;
		count -= val;
	}
	if(count)
		return false;
	return true;
}

bool
Tuner::Alsa::record(unsigned char *buf, int len) throw(Error::AudioError)
{
	int count = len;
	while(count >= info.fragSize) {
		int val = snd_pcm_plugin_read(handle, buf + len - count, info.fragSize);
		if(val <= 0)
			return false;
		count -= val;
	}

	if(count)
		return false;
	return true;
}

bool
Tuner::Alsa::avail(int *frags, int *bytes) throw(Error::AudioError)
{
	snd_pcm_channel_status_t status;

	memset(&status, 0, sizeof(status));
	status.channel = SND_PCM_CHANNEL_CAPTURE;
	status.mode = SND_PCM_MODE_BLOCK;

	if(snd_pcm_plugin_status(handle, &status) < 0)
		return false;

	*frags = status.count % info.fragSize;
	*bytes = status.count;

	return true;
}

bool
Tuner::Alsa::free(int *frags, int *bytes) throw(Error::AudioError)
{
	snd_pcm_channel_status_t status;

	memset(&status, 0, sizeof(status));
	status.channel = SND_PCM_CHANNEL_PLAYBACK;
	status.mode = SND_PCM_MODE_BLOCK;

	if(snd_pcm_plugin_status(handle, &status) < 0)
		return false;

	*frags = status.free % info.fragSize;
	*bytes = status.free;

	return true;
}

bool
Tuner::Alsa::isOpen() throw()
{
	return open;
}

bool
Tuner::Alsa::sync() throw(Error::AudioError)
{
	return true;
}

bool
Tuner::Alsa::stop() throw(Error::AudioError)
{
	if(info.mode == PLAYBACK || info.mode == DUPLEX)
		if(snd_pcm_plugin_playback_drain(handle) < 0)
			return false;
	if(info.mode == RECORD || info.mode == DUPLEX)
		if(snd_pcm_plugin_flush(handle, SND_PCM_CHANNEL_CAPTURE) < 0)
			return false;

	return true;
}

#endif

