/*  GuiTuner 0.05
 *  Copyright (c) 2001 Fabio Checconi <fchecconi@libero.it>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <cstring>
#include "error.h"
#include "i18n.h"
#include "calc.h"
#include "genconf.h"
#include "gui.h"
#include "mainthread.h"

void
Thread::MainThread::adjustBuffers(unsigned int av) throw()
{
	if(av < buffer->g_size()) {
		std::memmove(buffer->g_ptr(), buffer->g_ptr() + av, buffer->g_size() - av);
		std::memcpy(buffer->g_ptr() + buffer->g_size() - av, auxBuf.g_ptr(), av);
	} else {
		std::memcpy(buffer->g_ptr(), auxBuf.g_ptr() + av - buffer->g_size(),
			buffer->g_size());
	}
}

void
Thread::MainThread::openDevice(bool wait) throw()
{
	Tuner::AudioInfo *audio = fact.getAudioInfo();

	try {
		gui->freeze();
		if(!wait) {
			if(!audioDev->openDevice(audio, false, true)) {
				gui->errorMsg(_("Cannot open audio device, retry later"));
				paused = true;
				gui->thaw();
				return;
			}
		} else {
			bool open = false;
			for(int i = 0; i < 500 && !(open = audioDev->openDevice(audio, false, true)); ++i)
				gui->delay(1000);
			if(!open) {
				gui->errorMsg(_("Timed out opening audio device, retry later"));
				paused = true;
				gui->thaw();
				return;
			}
		}

		setup();
	} catch(std::exception &e) {
		gui->errorMsg(e.what());
		paused = true;
	}
	gui->thaw();
}

void
Thread::MainThread::closeDevice() throw()
{
	if(audioDev->isOpen())
		audioDev->closeDevice();
}

void
Thread::MainThread::setup() throw(std::bad_alloc)
{
	ai = *audioDev->getInfo();
	fact.setup(&ai);
	auxBuf.resize(ai.fragNumber * ai.fragSize);
	refreshBytes = static_cast<unsigned int>(refresh * ai.sampleSize() * ai.rate / 1000.0);
	if(refreshBytes >= auxBuf.g_size())
		refreshBytes = auxBuf.g_size() - ai.fragSize;
	refreshBytes -= refreshBytes % ai.fragSize;
}

Thread::MainThread::MainThread(Interface::Gui *g, Config::GenConf &gc, int ref)
	throw(Error::AudioError, Error::CalcError, Error::ThreadError, std::bad_alloc):
	fact(gc), gui(g), audioDev(fact.createAudio()), calc(fact.createCalc()),
	buffer(fact.createBuffer()), auxBuf(0), refresh(ref), refreshBytes(0), paused(false)
{
	ai.loadDefaults();
}

void
Thread::MainThread::start() throw(Error::ThreadError)
{
	if(!isRunning())
		run();
}

void
Thread::MainThread::stop() throw(Error::ThreadError)
{
	if(isRunning()) {
		Command cmd = STOP;
		sendMsg(&cmd, sizeof(Command));
		wait();
	}
}

void
Thread::MainThread::pause() throw(Error::ThreadError)
{
	Command cmd = PAUSE;
	if(isRunning())
		sendMsg(&cmd, sizeof(Command));
}

void
Thread::MainThread::resume() throw(Error::ThreadError)
{
	Command cmd = CONTINUE;
	if(isRunning())
		sendMsg(&cmd, sizeof(Command));
}

void
Thread::MainThread::reconfigure(int ref) throw()
{
	Command cmd = RECONFIGURE;
	sendMsg(&cmd, sizeof(cmd));
	sendMsg(&ref, sizeof(int));
	Response r;
	getMsg(&r, sizeof(r));
}

int
Thread::MainThread::main() throw()
{
	using namespace Tuner;

	int errorCount = 0;
	bool first = true;

	openDevice(false);

	while(errorCount < 2) {
		try {
			if(pending(FROM_PARENT)) {
				Command c;
				getMsg(&c, sizeof(Command), FROM_PARENT);
				switch(c) {
					case STOP:
						closeDevice();
						return errorCount;
					case PAUSE: {
						gui->pause();
						closeDevice();
						paused = first = true;
						break;
					}
					case CONTINUE: {
						paused = false;
						openDevice(true);
						gui->resume();
						break;
					}
					case RECONFIGURE: {
						int ref;
						getMsg(&ref, sizeof(int), FROM_PARENT);
						if(!paused) {
							gui->pause();
							closeDevice();
						}
						fact.update();
						refresh = ref;
						first = true;
						if(!paused) {
							openDevice(true);
							gui->resume();
						}
						Response r = DONE;
						sendMsg(&r, sizeof(r), TO_PARENT);
					}
				}
			}
			if(!paused) {
				int fn, b;
				unsigned int av;

				if(first) {
					std::memset(buffer->g_ptr(), 0, buffer->g_size());
					av = refreshBytes;
					first = false;
				} else {
					if(!audioDev->avail(&fn, &b))
						av = refreshBytes;
					else
						av = fn * ai.fragSize;
					if(av == 0 || av == auxBuf.g_size())
						av = auxBuf.g_size();
				}
				if(av < refreshBytes) {
					gui->delay(static_cast<unsigned int>(
						static_cast<double>(refreshBytes - av) *
						static_cast<double>(1000000) /
						static_cast<double>(ai.sampleSize()) /
						static_cast<double>(ai.rate)));
					continue;
				}

				if(!audioDev->record(auxBuf.g_ptr(), av)) {
					gui->errorMsg(std::string(
						_("Error while reading from audio device")));
					paused = true;
					continue;
				}
				adjustBuffers(av);
				gui->display(Tuner::Note((*calc)(buffer->g_ptr())));
			} else
				gui->delay(10000);
		} catch(std::exception &e) {
			gui->pause();
			gui->errorMsg(e.what());
			gui->resume();
			++errorCount;
		}
	}

	return -errorCount;
}

