/*  GuiTuner 0.05
 *  Copyright (c) 2001 Fabio Checconi <fchecconi@libero.it>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/*
 * NOTES: time-critical routines do NO strong error control, i.e. they do
 * not raise exceptions, this may be wrong.
 * The device name is not configurable at run time.
 * Cannot use cerrno, because errno can be a macro. 
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>
#include <cstring>
#include <string>
#include "oss.h"
#include "i18n.h"

namespace Tuner {
	const int BadFd = -1;

	const int DriverFrags = 0x7fff;
	const char DevName[] = "/dev/dsp";
}

void
Tuner::Oss::throwError(std::string s1, Error::AudioErrorID id, std::string s2)
	throw(Error::AudioError)
{
	if(fd != BadFd)
		::close(fd), fd = BadFd;
	throw Error::AudioError(s1, id, s2);
}

Tuner::Oss::Oss() throw(Error::AudioError): fd(BadFd)
{
		info.loadDefaults();
}

Tuner::Oss::~Oss()
{
	if(fd != BadFd)
		::close(fd);
}

bool
Tuner::Oss::openDevice(AudioInfo *ai, bool block, bool setBuf)
	throw(Error::AudioError)
{
	int flags = block ? 0: O_NONBLOCK;
	switch(info.mode = ai->mode) {
		case PLAYBACK:
			flags |= O_WRONLY;
			break;
		case RECORD:
			flags |= O_RDONLY;
			break;
		case DUPLEX:
			flags |= O_RDWR;
	}
	if((fd = open(DevName, flags)) == -1) {
		if(errno == EBUSY)
			return false;
		throwError(std::string("Tuner::Oss::openDevice"),
			Error::SYSTEM_CALL,
			std::string(std::strerror(errno)));
	}
	if(!block) {
		if((flags = fcntl(fd, F_GETFL)) == -1)
			throwError(std::string("Tuner::Oss::openDevice"),
				Error::SYSTEM_CALL,
				std::string(std::strerror(errno)));

		if((fcntl(fd, F_SETFL, flags & ~O_NONBLOCK)) == -1)
			throwError(std::string("Tuner::Oss::openDevice"),
				Error::SYSTEM_CALL,
				std::string(std::strerror(errno)));
	}
	if(!setFormat(&ai->fmt) || !setChannels(&ai->chn) ||
		!setRate(&ai->rate))
			throwError(std::string("Tuner::Oss::openDevice"),
				Error::SYSTEM_CALL,
				std::string(std::strerror(errno)));

	if(setBuf && ai->fragSize > 0) {
		ai->fragNumber = ai->fragNumber <= 0 ? DriverFrags:
			ai->fragNumber;
		if(!setBuffering(&ai->fragNumber, &ai->fragSize))
			throwError(std::string("Tuner::Oss::openDevice"),
				Error::SYSTEM_CALL,
				std::string(std::strerror(errno)));
	} else {
		if(!getBuffering(&ai->fragNumber, &ai->fragSize))
			throwError(std::string("Tuner::Oss::openDevice"),
				Error::SYSTEM_CALL,
				std::string(std::strerror(errno)));
	}

	info = *ai;

	return true;
}

bool
Tuner::Oss::closeDevice() throw()
{
	::close(fd), fd = BadFd;

	return true;
}

bool
Tuner::Oss::play(unsigned char *buf, int len) throw(Error::AudioError)
{
	return ::write(fd, buf, len) != -1;
}

bool
Tuner::Oss::record(unsigned char *buf, int len) throw(Error::AudioError)
{
	return ::read(fd, buf, len) != -1;
}

bool
Tuner::Oss::avail(int *frags, int *bytes) throw(Error::AudioError)
{
	audio_buf_info info;

	if(ioctl(fd, SNDCTL_DSP_GETISPACE, &info) == -1)
		return false;

	*frags = info.fragments;
	*bytes = info.bytes;

	return true;
}

bool
Tuner::Oss::free(int *frags, int *bytes) throw(Error::AudioError)
{
	audio_buf_info info;

	if(ioctl(fd, SNDCTL_DSP_GETOSPACE, &info) == -1)
		return false;

	*frags = info.fragments;
	*bytes = info.bytes;

	return true;
}

bool
Tuner::Oss::isOpen() throw()
{
	return fd != BadFd;
}

bool
Tuner::Oss::sync() throw(Error::AudioError)
{
	if(ioctl(fd, SNDCTL_DSP_SYNC, 0) == -1)
		return false;

	return true;
}

bool
Tuner::Oss::stop() throw(Error::AudioError)
{
	if(ioctl(fd, SNDCTL_DSP_RESET, 0) == -1)
		return false;

	return true;
}

bool
Tuner::Oss::setFormat(Format *f)
{
	int arg;

	switch(*f) {
		case U8:
			arg = AFMT_U8;
			break;
		case S16BE:
			arg = AFMT_S16_BE;
			break;
		case S16LE:
			arg = AFMT_S16_LE;
	}

	if(ioctl(fd, SNDCTL_DSP_SETFMT, &arg) == -1)
		return false;

	switch(arg) {
		case AFMT_U8:
			*f = U8;
			break;
		case AFMT_S16_BE:
			*f = S16BE;
			break;
		case AFMT_S16_LE:
			*f = S16LE;
			break;
		default:
			throwError(std::string("Tuner::Oss::openDevice"),
				Error::SYSTEM_CALL,
				std::string(_("This OSS driver doesn' t support any compatible format")));
	}

	return true;
}

bool
Tuner::Oss::setChannels(Channels *c)
{
	int arg;

	switch(*c) {
		case MONO:
			arg = 1;
			break;
		case STEREO:
			arg = 2;
	}

	if(ioctl(fd, SNDCTL_DSP_CHANNELS, &arg) == -1)
		return false;

	switch(arg) {
		case 1:
			*c = MONO;
			break;
		case 2:
			*c = STEREO;
			break;
		default:
			throwError(std::string("Tuner::Oss::openDevice"),
				Error::SYSTEM_CALL,
				std::string(_("This OSS driver doesn' t support any compatible format")));;
	}

	return true;
}

bool
Tuner::Oss::setRate(SamplingRate *r)
{
	if(ioctl(fd, SNDCTL_DSP_SPEED, static_cast<int *>(r)) == -1)
		return false;

	return true;
}

bool
Tuner::Oss::setBuffering(int *n, int *s)
{
	int arg = ((*n)&0xffff) << 16, exp = 0;

	while( *s > 1 ) {
		*s >>= 1;
		exp++;
	}

	arg |= (exp&0xffff);
	if(ioctl(fd, SNDCTL_DSP_SETFRAGMENT, &arg) == -1)
		return false;

	if(!getBuffering(n, s))
		return false;

	return true;
}

bool
Tuner::Oss::getBuffering(int *n, int *s)
{
	audio_buf_info dev_info;

	switch(info.mode) {
		case PLAYBACK:
		case DUPLEX:
			if(ioctl(fd, SNDCTL_DSP_GETOSPACE, &dev_info) == -1)
				return false;
			break;
		case RECORD:
			if(ioctl(fd, SNDCTL_DSP_GETISPACE, &dev_info) == -1)
				return false;
	}

	*n = dev_info.fragstotal;
	*s = dev_info.fragsize;

	return true;
}

